package org.dasein.cloud.jclouds.rackspace;

import static org.jclouds.concurrent.MoreExecutors.sameThreadExecutor;

import java.util.concurrent.Callable;

import org.apache.log4j.Logger;
import org.dasein.cloud.AbstractCloud;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.services.access.AccessServices;
import org.dasein.cloud.services.accounting.AccountingServices;
import org.dasein.cloud.services.address.AddressServices;
import org.dasein.cloud.services.cdn.ContentDistributionServices;
import org.dasein.cloud.services.dc.DataCenterServices;
import org.dasein.cloud.services.disk.SnapshotServices;
import org.dasein.cloud.services.disk.VolumeServices;
import org.dasein.cloud.services.firewall.FirewallServices;
import org.dasein.cloud.services.image.ImageServices;
import org.dasein.cloud.services.lb.LoadBalancerServices;
import org.dasein.cloud.services.scaling.ScalingServices;
import org.dasein.cloud.services.server.ServerServices;
import org.dasein.cloud.services.storage.StorageServices;
import org.dasein.util.Retry;
import org.jclouds.blobstore.BlobStoreContextFactory;
import org.jclouds.compute.ComputeServiceContextFactory;
import org.jclouds.concurrent.config.ExecutorServiceModule;
import org.jclouds.logging.log4j.config.Log4JLoggingModule;
import org.jclouds.rackspace.cloudfiles.CloudFilesAsyncClient;
import org.jclouds.rackspace.cloudfiles.CloudFilesClient;
import org.jclouds.rackspace.cloudservers.CloudServersAsyncClient;
import org.jclouds.rackspace.cloudservers.CloudServersClient;
import org.jclouds.rest.RestContext;

import com.google.common.collect.ImmutableSet;

public class RackspaceCloud extends AbstractCloud {

	static private final Logger logger = Logger.getLogger(RackspaceCloud.class);
	
	@Override
	public AccessServices getAccessServices() {
		// TODO Auto-generated method stub
		return null;
	}

	private transient AccountingServices accountingServices = null;
	
	@Override
	public AccountingServices getAccountingServices() {
	    if( accountingServices == null ) {
	        accountingServices = new RackspacePricing();
	    }
	    return accountingServices;
	}
	
	private transient AddressServices addressServices = null;

	@Override
	public AddressServices getAddressServices() {
		if( addressServices == null ) {
			addressServices = new RackspaceAddress(this);
        }
        return addressServices;
	}

	private transient ContentDistributionServices cdnServices = null;
	
	@Override
	public ContentDistributionServices getCdnServices() {
		if( cdnServices == null ) {
			cdnServices = new RackspaceCDN(this);
        }
        return cdnServices;
	}

	@Override
	public String getCloudName() {
		return "Cloud Servers";
	}

	public RestContext<CloudFilesClient,CloudFilesAsyncClient> getCloudFilesContext() {
        logger.debug("enter - getCloudFilesContext()");
        try {
            ProviderContext ctx = getContext();
            String user = new String(ctx.getAccessPublic());
            String key = new String(ctx.getAccessPrivate());
            return new BlobStoreContextFactory().createContext("cloudfiles", user, key, ImmutableSet.of(new Log4JLoggingModule(), new ExecutorServiceModule(sameThreadExecutor(), sameThreadExecutor()))).getProviderSpecificContext();
        }
        finally {
            logger.debug("exit - getCloudFilesContext()");
        }	    
	}
	
    public RestContext<CloudServersClient,CloudServersAsyncClient> getCloudServersContext() throws CloudException {
        logger.debug("enter - getCloudServersContext()");
        try {
            return (new Retry<RestContext<CloudServersClient,CloudServersAsyncClient>>()).retry(5, new Callable<RestContext<CloudServersClient,CloudServersAsyncClient>>() {
                @Override
                public RestContext<CloudServersClient,CloudServersAsyncClient> call() throws Exception {
                    ProviderContext ctx = getContext();
                    String user = new String(ctx.getAccessPublic());
                    String key = new String(ctx.getAccessPrivate());
                    
                    return new ComputeServiceContextFactory().createContext("cloudservers", user, key, ImmutableSet.of(new Log4JLoggingModule(), new ExecutorServiceModule(sameThreadExecutor(), sameThreadExecutor()))).getProviderSpecificContext();
                }
            });
        }
        catch( Exception e ) {
            e.printStackTrace();
            throw new CloudException("Unable to connect to cloud: " + e.getMessage());
        }
        finally {
            logger.debug("exit - getCloudServersContext()");
        }       
    }
    
    private transient DataCenterServices dataCenterServices = null;
    
	@Override
	public DataCenterServices getDataCenterServices() {
	    if( dataCenterServices == null ){
	        dataCenterServices = new RackspaceDataCenter(this);
        }
        return dataCenterServices;
	}

	@Override
	public FirewallServices getFirewallServices() {
		// TODO Auto-generated method stub
		return null;
	}

	private transient ImageServices imageServices = null;
    
    @Override
    public synchronized ImageServices getImageServices() {
        if( imageServices == null ) {
            imageServices = new RackspaceImage(this);
        }
        return imageServices;
    }

	@Override
	public LoadBalancerServices getLoadBalancerServices() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String getProviderName() {
		return "Rackspace";
	}

	@Override
	public ScalingServices getScalingServices() {
		// TODO Auto-generated method stub
		return null;
	}

	private transient ServerServices serverServices = null;
    
    @Override
    public synchronized ServerServices getServerServices() {
        if( serverServices == null ) {
            serverServices = new RackspaceServer(this);
        }
        return serverServices;
    }

	@Override
	public SnapshotServices getSnapshotServices() {
		// TODO Auto-generated method stub
		return null;
	}

	private transient StorageServices storageServices = null;
	
	@Override
	public StorageServices getStorageServices() {
		if( storageServices == null ) {
			storageServices = new RackspaceStorage(this);
        }
        return storageServices;
	}

	@Override
	public VolumeServices getVolumeServices() {
		// TODO Auto-generated method stub
		return null;
	}
}
