package org.dasein.cloud.jclouds.rackspace;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.image.ImageServices;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Architecture;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.storage.CloudStoreObject;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;
import org.jclouds.rackspace.cloudservers.CloudServersAsyncClient;
import org.jclouds.rackspace.cloudservers.CloudServersClient;
import org.jclouds.rackspace.cloudservers.domain.Image;
import org.jclouds.rackspace.cloudservers.options.ListOptions;
import org.jclouds.rest.RestContext;

public class RackspaceImage implements ImageServices {

	static private final Logger logger = Logger.getLogger(RackspaceImage.class);
    
    private RackspaceCloud provider = null;
    
    public RackspaceImage(RackspaceCloud provider) {
        this.provider = provider;
    }
	
	public String create(String location) throws InternalException, CloudException {
		throw new OperationNotSupportedException();
	}

	public void delete(String imageId) throws InternalException, CloudException {
	    logger.debug("enter - delete(String)");
	    try {
	        // TODO: implement me
	    }
	    finally {
	        logger.debug("exit - delete(String)");
	    }
	}

	public Collection<ServerImage> find(String executableByAccountNumber, String keyword, Platform platform, String architecture) throws InternalException,	CloudException {
	    logger.debug("enter - find(String, String, Platform, String)");
	    try {
    		Collection<ServerImage> images = list(executableByAccountNumber);
    		Collection<ServerImage> found = new ArrayList<ServerImage>();
    		
    		for (ServerImage image : images) {
    			if (image.getArchitecture().name().equals(architecture) && image.getName().indexOf(keyword) > -1 && image.getPlatform().equals(platform)) {
    				found.add(image);
    			}
    		}
    		
    		return found;
	    }
	    finally {
	        logger.debug("exit - find(String, String, Platform, String)");
	    }
	}

	private CloudServersAsyncClient getAsyncClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
	    return ctx.getAsyncApi();
	}
	
    private CloudServersClient getClient(RestContext<CloudServersClient,CloudServersAsyncClient> ctx) {
        return ctx.getApi();
    }
    
	public ServerImage getImage(String imageId) throws InternalException, CloudException {
	    for( ServerImage img : list(provider.getContext().getAccountNumber()) ) {
	        if( img.getProviderServerImageId().equals(imageId) ) {
	            return img;
	        }
	    }
	    return null;
	}

	public CloudStoreObject getManifest(String imageId) throws CloudException, InternalException {
		return null;
	}

	public String getProviderTermForImage(Locale locale) {
		return "image";
	}

	public AsynchronousTask<ServerImage> imageServer(String serverId, String forAccountAccessKey, String directory, String imageName) throws InternalException, CloudException {
		logger.debug("enter - imageServer(String, String, String, String)");
		try {
		    final AsynchronousTask<ServerImage> task = new AsynchronousTask<ServerImage>();
            final Future<Image> result;

            task.setStartTime(System.currentTimeMillis());

            RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

            try {
                result = getAsyncClient(ctx).createImageFromServer(imageName, Integer.parseInt(serverId));
            }
            finally {
                ctx.close();
            }
		    Thread t = new Thread() {
		        public void run() {
		            while( !result.isDone() ) {
		                try { Thread.sleep(5000L); }
		                catch( InterruptedException e ) { }
		            }
		            task.setPercentComplete(100.0);
		            if( result.isCancelled() ) {
		                task.complete(new InterruptedException("Bundle Canceled"));
		            }
		            else {
		                try {
		                    Image img = result.get();
		                    
		                    task.completeWithResult(toImage(img));
		                }
		                catch( Exception e ) {
		                    logger.error("Bundle of server failed: " + e.getMessage());
		                    e.printStackTrace();
		                    task.complete(new CloudException(e));
		                }
		            }
		        }
		    };
		    
		    t.setDaemon(true);
		    t.setName("Create Image - " + imageName + "/" + serverId);
		    t.start();
		    return task;
		}
		finally {
		    logger.debug("exit - imageServer(String, String, String, String)");
		}
	}

	public boolean isSupportsCustomImages() {
	    return true;
	}
	
    public boolean isSupportsSharing() {
        return false;
    }
    
	public Collection<ServerImage> list(String accountNumber) throws InternalException, CloudException {
        PopulatorThread<ServerImage> populator;
        final String acct = accountNumber;
        
        provider.hold();
        populator = new PopulatorThread<ServerImage>(new JiteratorPopulator<ServerImage>() {
            public void populate(Jiterator<ServerImage> iterator) throws CloudException, InternalException {
                try {
                    populateImages(acct, iterator);
                }
                finally {
                    provider.release();
                }
            }
        });
        populator.populate();
        return populator.getResult();
	}
	
	private void populateImages(String accountNumber, Jiterator<ServerImage> iterator) throws CloudException, InternalException {
	    logger.debug("enter - list(String)");
	    try {
	        Collection<Image> rawImages;
	        
	        try {
	            RestContext<CloudServersClient,CloudServersAsyncClient> ctx = provider.getCloudServersContext();

	            try {
	                rawImages = getClient(ctx).listImages(ListOptions.Builder.withDetails());
	            }
	            finally {
	                ctx.close();
	            }
	        }
	        catch( RuntimeException e ) {
	            logger.error("Could not list images from Rackspace: " + e.getMessage());
	            e.printStackTrace();
	            throw new CloudException(e);
	        }
	        for( Image image: rawImages ) {
	            if( image != null ) {
	                iterator.push(toImage(image));
	            }
	        }
	    }
	    finally {
	        logger.debug("exit - list(String)");
	    }
	}

    public Collection<String> getShares(String imageId) throws CloudException, InternalException {
        throw new OperationNotSupportedException();
    }

    public boolean isPublic(String imageId) throws InternalException, CloudException {
        return false;
    }

    public void share(String imageId, String withAccountId, boolean allowed) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }		
    
    private ServerImage toImage(Image rackspaceImage) throws InternalException, CloudException {
        ServerImage image = new ServerImage();

        image.setOwnerAccountId(provider.getContext().getAccountNumber());
        image.setPersistent(true);
        image.setPublicImage(false);
        image.setType("standard");

        image.setName(rackspaceImage.getName());
        image.setDescription(image.getName());
        image.setArchitecture(Architecture.I64);
        image.setPlatform(Platform.guess(image.getName()));
        if( image.getPlatform().equals(Platform.UNKNOWN) ) {
            image.setPlatform(Platform.UNIX);
        }
        image.setProviderServerImageId(String.valueOf(rackspaceImage.getId()));
        
        RackspaceServer svc = (RackspaceServer)provider.getServerServices();
        
        image.setSize(svc.getSixtyFours());
        switch( rackspaceImage.getStatus() ) {
            case ACTIVE: 
                image.setActive(true);
                image.setAvailable(true);
                break;
            case PREPARING: case QUEUED: case SAVING: case UNRECOGNIZED:
                image.setActive(true);
                image.setAvailable(false);
                break;
            case FAILED:
                image.setActive(false);
                image.setAvailable(false);
                break;
            default:
                image.setActive(true);
                image.setAvailable(false);
                break;
        }
        return image;
    }
}
